/* Simple MPI exchange to a set of "neighbours" (other nodes) for latency
 * and bandwidth measurement.
 * Usage:
 *  mpi_latency NFLOATS ITERATIONS OVERLAP NEIGHBOURS DIRNAME
 * NFLOATS is number of floats to send/receive in each message
 * ITERATIONS is the number of exchange rounds on each process
 * OVERLAP of 0 causes blocking send/receive to all neighbours.
 * If 1, performs non-blocking Isend/Ireceive to all neighbours.
 * If 2, also performs a barrier after all non-blocking Isend/Ireceives.
 * If 3, performs a reduceall instead of a barrier.
 * (OVERLAP of 1 should be twice as fast as 0 on most systems, barrier/reduce
 * slower but a realistic variant of many of our timestep loops)
 * NEIGHBOURS is number of neighbours to exchange messages with.
 * Must be at least 1,* less than MAX_NEIGHBOURS,
 * must be less than half of the number of nodes
 * DIRNAME is the name of the directory to dump statistics into. Filenames are
 * DIRNAME/OVERLAP_NEIGHBOURS_PROCESSES_FLOATS and
 * DIRNAME/OVERLAP_NEIGHBOURS_PROCESSES_FLOATS.sum (the former contains the
 * histogram of timings for exchanges on every process, the latter
 * contains the sum of all histograms)
 */
/* Mark Moraes & Kevin J. Bowers, D. E. Shaw Research, LLC */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <signal.h>
#include <mpi.h>
#include "mpi_err.h"

#include "hrtimer.h"

/* increase this if you want to run on more than 2048 MPI processes */
#define MAX_NEIGHBOURS	1024

/*
 * logarithmically sized bins to collect statistics about how long each
 * exchange takes using data from the high resolution timer, 100 bins
 * per decade, time converted to nsec
 */
#define NBINS		1500
#define BIN(x)		(int) (log10((x)*1000/cycles_per_usec)*100.0)
#define USEC(bin)	( (exp10((bin+1)/100.0) + exp10((bin+1)/100.0))/2.0/1000.0 )

const char *overlap_names[] = {
    "No", "Yes", "Barrier", "Allreduce",
};

int debug = 0;
int rank = -1;
unsigned int xtimeout;


void
timeout(int i)
{
    error(13, "Timeout after %u seconds\n", xtimeout);
    /* NOTREACHED */
}

void
usage(int exitstatus, const char *fmt, ...)
{
  va_list ap;
  if (rank == 0) {
      fprintf(stderr,
	"Usage: mpi_latency nfloats iterations overlap neighbours odirname\n"
	"where overlap is 0 for non-overlapped send/recv instead of Isend/Irecv\n"
	"                 1 for barrier every 1000 iterations\n"
	"                 2 for barrier every iteration\n"
	"                 3 for allreduce every iteration\n");
  }
  va_start(ap, fmt);
  verror(exitstatus, fmt, ap);
  va_end(ap); /* not reached */
}

void
overlap_exchange(int id, int n, int nneighbours,
		 float *msgsend,	/* [n] */
		 int *dst,		/* [nneighbours] */
		 float **msgrecv,	/* [nneighbours][n] */
		 int *src		/* [n] */) {
  MPI_Request mreq[MAX_NEIGHBOURS*2];
  MPI_Status mstat[MAX_NEIGHBOURS*2];
  int i, j, e;

  /*
   * for async, we always post our receive first so that MPI has the
   * buffer if it wishes to optimize by receiving straight into the buffer.
   */
  for (i = 0; i < nneighbours; i++) {
    if ( debug )
      printf("iter %d rank %d exch recving from %d\n", id, rank, src[i]);
    if ((e = MPI_Irecv( msgrecv[i], n, MPI_FLOAT, src[i], id,
		        MPI_COMM_WORLD, &mreq[i])) != MPI_SUCCESS)
      error(21, "overlap_exchange Irecv from %d failed: %d", src[i], e);
  }
  for (i = 0, j = nneighbours; i < nneighbours; i++, j++) {
    if ( debug )
      printf("iter %d rank %d exch sending to %d\n", id, rank, dst[i]);
    if ((e = MPI_Isend( msgsend, n, MPI_FLOAT, dst[i], id,
		        MPI_COMM_WORLD, &mreq[j])) != MPI_SUCCESS)
      error(23, "overlap_exchange Isend to %d failed: %d", dst[i], e);
  }
  if ( debug )
    printf("iter %d rank %d waiting for %d pending ops\n", id, rank,
	   nneighbours*2);
  if ((e = MPI_Waitall(nneighbours*2, mreq, mstat)) != MPI_SUCCESS)
      error(25, "overlap_exchange Waitall failed: %d", e);
  if ( debug )
    printf("iter %d rank %d exch done\n", id, rank);
}

void
ping(int id, int n, int nneighbours,
     float *msgsend,	/* [n] */
     int *dst,		/* [nneighbours] */
     float **msgrecv,	/* [nneighbours][n] */
     int *src		/* [n] */) {
  /* Note: all dst[i] and src[i] neighbours must be running pong() */
  int i, e;
  for (i = 0; i < nneighbours; i++) {
    if ( debug )
      printf("iter %d rank %d ping sending to %d\n", id, rank, dst[i]);
    if ((e = MPI_Send( msgsend, n, MPI_FLOAT, dst[i], 2*id,
		       MPI_COMM_WORLD )) != MPI_SUCCESS)
      error(31, "ping Send to %d failed: %d", dst[i], e);
  }
  for (i = 0; i < nneighbours; i++) {
    if ( debug )
      printf("iter %d rank %d ping recving from %d\n", id, rank, src[i]);
    if ((e = MPI_Recv( msgrecv[i], n, MPI_FLOAT, src[i], 2*id+1,
		      MPI_COMM_WORLD, MPI_STATUS_IGNORE )) != MPI_SUCCESS)
      error(33, "ping Recv from %d failed: %d", src[i], e);
  }
}

void
pong(int id, int n, int nneighbours,
     float *msgsend,	/* [n] */
     int *dst,		/* [nneighbours] */
     float **msgrecv,	/* [nneighbours][n] */
     int *src		/* [n] */) {
  /* Note: all dst[i] and src[i] neighbours must be running ping() */
  int i, e;
  for (i = 0; i < nneighbours; i++) {
    if ( debug )
      printf("iter %d rank %d pong recving from %d\n", id, rank, src[i]);
    if ((e = MPI_Recv( msgrecv[i], n, MPI_FLOAT, src[i], 2*id,
		      MPI_COMM_WORLD, MPI_STATUS_IGNORE )) != MPI_SUCCESS)
      error(41, "ping Recv from %d failed: %d", src[i], e);
  }
  for (i = 0; i < nneighbours; i++) {
    if ( debug )
      printf("iter %d rank %d pong sending to %d\n", id, rank, dst[i]);
    if ((e = MPI_Send( msgsend, n, MPI_FLOAT, dst[i], 2*id+1,
		       MPI_COMM_WORLD )) != MPI_SUCCESS)
      error(43, "ping Send to %d failed: %d", dst[i], e);
  }
}

void
accum_bins(unsigned long *bins, int n, unsigned long *sumbins)
{
  int i;
  for (i = 0; i < n; i++)
    sumbins[i] += bins[i];

}

/* write statistics bins to a file */
void
dump_bins(FILE *fp, unsigned long *bins, int n, int binrank)
{
  int i;
  if (!fp)
    return;
  fprintf(fp, "# stats from rank %d\n", binrank);
  for (i = 0; i < n; i++) {
    if (bins[i] > 0 || (i > 0 && bins[i-1] > 0)) {
      fprintf(fp, "%g %d\n", USEC(i), bins[i]);
    }
  }
  fprintf(fp, "\n\n");
  if (ferror(fp) || (fflush(fp) != 0)) {
    error(9, "error saving bin %s");
  }
}

/*
 * Just a kludge multiplier used when estimating a watchdog alarm to
 * avoid hangs
 */
double ovfactor[] = { 3, 20, 2.75, 3 };

int
main( int argc,
      char ** argv ) {
  float * msgsend, * msgrecv[MAX_NEIGHBOURS];
  int i, j, e, n, r, size, overlap, nneighbours, bsize, warmup, maxiter;
  int *maxtag;
  int sneighbours[MAX_NEIGHBOURS], rneighbours[MAX_NEIGHBOURS];
  unsigned long *hrbins, *sumhrbins, *hrbintmp;
  unsigned long binunderflow = 0, binoverflow = 0;
  char fnamebuf[PATH_MAX]; /* XXX */
  char *odirname, *cp;
  double elapsed;
  HRTIME_T cycles_per_usec, hr1, hr2;

  if ((cp = getenv("MPI_LATENCY_DEBUG")) != NULL)
    debug = atoi(cp);
  if ((e = MPI_Init( &argc, &argv )) != MPI_SUCCESS)
    error(11, "MPI_Init failed: %d", e);
  if ((e = MPI_Comm_size( MPI_COMM_WORLD, &size )) != MPI_SUCCESS)
    error(13, "MPI_Comm_size failed: %d", e);
  if ((e = MPI_Comm_rank( MPI_COMM_WORLD, &rank )) != MPI_SUCCESS)
    error(14, "MPI_Comm_Rank failed: %d", e);
  if ((e = MPI_Attr_get( MPI_COMM_WORLD, MPI_TAG_UB, (void *)&maxtag, &i))
      != MPI_SUCCESS)
    error(15, "MPI_Attr_get MPI_TAG_UB failed: %d\n", e);
  if( argc != 6 )
    usage(1, "Wrong number of arguments");
  if( (size%2) != 0 || size < 2)
    usage(2, "Need even number of processors");
  if ((n = atoi( argv[1] )) <= 0 || n < 1)
    usage(3, "nfloats must be a natural number");
  if ((r = atoi( argv[2] )) <= 0 || r < 1)
    usage(4, "iterations must be a natural number");
  overlap = atoi ( argv[3] );
  if ( overlap < 0 || overlap > 3) usage(8, "overlap must be 0..3");
  nneighbours = atoi( argv[4] );
  i = overlap ? *maxtag : *maxtag/2;
  if (r > i)
    usage(4, "max tag is %d, so max iterations must be less than %d", *maxtag, i);
  i = size/2 < MAX_NEIGHBOURS ? size/2 : MAX_NEIGHBOURS;
  if (nneighbours < 1 || nneighbours > i)
    usage(7, "neighbours must be 1..%d", i);
  odirname = argv[5];

  /*
   * we have occasionally seen fabrics or MPI hang under load, this is
   * just a defensive measure so that the run exits rather than sits
   * around forever
   */
  elapsed = 5 + nneighbours*r*(50+(n*sizeof(float)/200.0))*1e-6*ovfactor[overlap]*log(size); /* XXX */
  if (debug && rank == 0)
    printf("timeout %g\n", elapsed);
  xtimeout = 10*elapsed;
  signal(SIGALRM, timeout);
  alarm(xtimeout);

  /* allocate and initialize buffers */
  bsize = n*sizeof(float);
  if ((msgsend = (float *)malloc( bsize )) == NULL )
    error(5, "malloc %u failed", bsize);
  for (i = 0; i < nneighbours; i++) {
    float *fp;
    
    fp = (float *)malloc( bsize );
    if( fp ==NULL )
	error(5, "malloc %u failed", bsize);
    for (j = 0; j < n; j++)
      fp[j] = 0;
    msgrecv[i] = fp;
  }
  for( i=0; i<n; i++ )
    msgsend[i] = i*r;

  if (rank == 0) {
    printf( "Size (of each message)           = %i floats/message\n"
	    "Count (iterations)               = %i messages\n"
            "Number of CPUs                   = %i\n"
	    "Number of Neighbours             = %i\n"
	    "Should overlap all send/recv     = %s\n",
	    n,
	    r,
	    size,
	    nneighbours,
	    overlap_names[overlap] );
    fflush(stdout);
  }

  /*
   * figure out my neighbours.  Just to keep things simple, we just
   * increment rather than go in both directions.
   */
  for ( i = 0; i < nneighbours; i++ ) {
    sneighbours[i] = (rank + i*2 + 1) % size;
    rneighbours[i] = ((rank - i*2 - 1) + size) % size;
    if ( debug ) {
      printf("%d: %d %d\n", rank, sneighbours[i], rneighbours[i]);
    }
  }
    

  /* Calibrate */
  cycles_per_usec = HRTIMER_USEC();

  if (cycles_per_usec) {
    if ((hrbins = calloc(NBINS, sizeof(unsigned long))) == NULL)
	error(6, "calloc( %u %u ) hrbins failed", NBINS, sizeof(unsigned long));
    if (rank == 0) {
	if ((sumhrbins = calloc(NBINS, sizeof(unsigned long))) == NULL)
	    error(6, "calloc( %u %u ) sumhrbins failed", NBINS, sizeof(unsigned long));
	if ((hrbintmp = calloc(NBINS, sizeof(unsigned long))) == NULL)
	    error(6, "calloc( %u %u ) hrbintmp failed", NBINS, sizeof(unsigned long));
    } else {
	sumhrbins = hrbintmp = NULL;
    }
  }
    
  /* make sure everyone is ready to go */
  if ((e = MPI_Barrier( MPI_COMM_WORLD )) != MPI_SUCCESS)
    error(17, "initial MPI_Barrier failed: %d", e);

  if (debug && rank == 0)
    printf("Barrier done\n");
  warmup = 1000;
  maxiter = r + warmup;
  for (i = 0; i < maxiter; i++) {
    hr1 = HRTIMER();
    if (warmup > 0) {
	warmup--;
    } else if (warmup == 0) {
	if (( e = MPI_Barrier( MPI_COMM_WORLD )) != MPI_SUCCESS) {
	    error(18, "MPI_Barrier after warmup failed: %d\n", e);
	}
	if (debug && rank == 0) {
	    printf("warmup finished\n");
	}
	warmup = -1;
	elapsed = MPI_Wtime();
    }
    if (overlap) {
      overlap_exchange( i, n, nneighbours, msgsend, sneighbours,
			msgrecv, rneighbours);
      if(overlap == 2 || (i > 0 && (i%1000) == 0)) {
	if ((e = MPI_Barrier( MPI_COMM_WORLD )) != MPI_SUCCESS)
	  error(18, "MPI_Barrier failed after %d iterations: %d", i, e);
	if (debug && rank == 0)
	    printf("%d: Barrier done\n", i);
      } else if( overlap == 3 ) {
	if ((e = MPI_Allreduce ( msgrecv[0], msgsend, n, MPI_FLOAT, MPI_SUM,
		       MPI_COMM_WORLD )) != MPI_SUCCESS)
	  error(19, "MPI_Allreduce failed after %d iterations: %d", i, e);
	/* scale reduced value back, so it doesn't blow up to NaN eventually */
	for (j = 0; j < n; j++) {
	    if (debug)
		printf("%d %d %d %g\n", rank, i, j, msgsend[j]);
	    msgsend[j] /= size;
	}
      }
    } else if( (rank%2) ==0 ) {
      /*
       * even numbered nodes ping: first send to sneighbours, then receive
       * from rneighbours
       */
      ping( i, n, nneighbours, msgsend, sneighbours, msgrecv, rneighbours);
    } else {
     /*
      * odd numbered nodes pong: first receive from rneighbours, then send
      * to sneighbours
      */
      pong( i, n, nneighbours, msgsend, sneighbours, msgrecv, rneighbours);
    }
    hr2 = HRTIMER();
    if (warmup < 0 && cycles_per_usec) {
      int nbin = BIN(hr2 - hr1);
      if (nbin < 0 ) {
	binunderflow++;
      } else if (nbin >= NBINS) {
	binoverflow++;
      } else {
	hrbins[nbin]++;
      }
    }
  }
  /* make sure everyone is done */
  (void) MPI_Barrier( MPI_COMM_WORLD );
  elapsed = MPI_Wtime() - elapsed;

  if( rank==0 ) {
    double MBsent = 1.*n*r*sizeof(float)/1024./1024.;
    printf( "Data sent/CPU/neighbour          = %g MB\n"
	    "Data sent/CPU                    = %g MB\n"
	    "Total data sent                  = %g MB\n"
	    "Total time                       = %g s\n"
	    "Avg round-trip time/iteration    = %g us\n"
	    "Avg send bandwidth/CPU/neighbour = %g MB/s\n"
	    "Avg send bandwidth/CPU           = %g MB/s\n"
	    "Total send bandwidth             = %g MB/s\n",
            MBsent,
	    MBsent*nneighbours,
	    MBsent*nneighbours*size,
	    elapsed,
	    elapsed*1e6/(double)(r),
	    MBsent/elapsed,
	    MBsent*nneighbours/elapsed,
	    MBsent*nneighbours*size/elapsed );
  }
  fflush(stdout);

  if (cycles_per_usec) {
    if (rank != 0) {
	(void) MPI_Send(hrbins, NBINS, MPI_UNSIGNED_LONG, 0, 0,MPI_COMM_WORLD);
    } else {
      unsigned int hrbinmin = NBINS, hrbinmax = 0, hrbinmode = 0;
      unsigned long hrmodeval = 0;
      FILE *fp = NULL;
      if (*odirname && strcmp(odirname, "-") != 0) {
	(void) snprintf(fnamebuf, sizeof(fnamebuf)-1, "%s/%d_%d_%05d_%06d",
			odirname, overlap, nneighbours, size, n);
	if ((fp = fopen(fnamebuf, "wc")) == NULL)
	    error(8, "error opening file %s for writing", fnamebuf);
	fprintf(fp, "#>>> %d processors: floats=%d iterations=%d overlap=%d neighbours=%d\n",
		size, n, r, overlap, nneighbours);
      }
      accum_bins(hrbins, NBINS, sumhrbins);
      dump_bins(fp, hrbins, NBINS, 0);
      for (i = 1; i < size; i++) {
	/* get everyone else's hrbins */
	(void) MPI_Recv(hrbintmp, NBINS, MPI_UNSIGNED_LONG, MPI_ANY_SOURCE,
			MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	accum_bins(hrbintmp, NBINS, sumhrbins);
	dump_bins(fp, hrbintmp, NBINS, i);
      }
      if (fp) {
	(void) fclose(fp);
	(void) snprintf(fnamebuf, sizeof(fnamebuf)-1, "%s/%d_%d_%05d_%06d.sum",
			odirname, overlap, nneighbours, size, n);
	if ((fp = fopen(fnamebuf, "wc")) == NULL)
	    error(8, "error opening file %s for writing", fnamebuf);
	fprintf(fp, "#>>> %d processors: floats=%d iterations=%d overlap=%d neighbours=%d\n",
		size, n, r, overlap, nneighbours);
	dump_bins(fp, sumhrbins, NBINS, -1);
	(void) fclose(fp);
      }
      for (i = 0; i < NBINS; i++) {
	if ( debug && sumhrbins[i])
	  printf("%d %lu\n", i, sumhrbins[i]);
	if (sumhrbins[i] > 0 && i < hrbinmin)
	  hrbinmin = i;
	if (sumhrbins[i] > 0 && i > hrbinmax)
	  hrbinmax = i;
	if (sumhrbins[i] > sumhrbins[hrbinmode]) {
	  hrbinmode = i;
	}
      }
      printf("Min us / bin index / bin count   = %g / %u / %lu\n"
             "Mode us / bin index / bin count  = %g / %u / %lu\n"
             "Max us / bin index / bin count   = %g / %u / %lu\n",
	     USEC(hrbinmin), hrbinmin, sumhrbins[hrbinmin],
	     USEC(hrbinmode), hrbinmode, sumhrbins[hrbinmode],
	     USEC(hrbinmax), hrbinmax, sumhrbins[hrbinmax]);
      fflush(stdout);
    }
    free( hrbins );
    if (rank == 0) {
      free( sumhrbins );
      free( hrbintmp );
    }
  }

  free( msgsend );
  for (i = 0; i < nneighbours; i++)
    free( msgrecv[i] );
  MPI_Finalize();
  return 0;
}
